'use strict'
var __createBinding =
	(this && this.__createBinding) ||
	(Object.create
		? function (o, m, k, k2) {
				if (k2 === undefined) k2 = k
				Object.defineProperty(o, k2, {
					enumerable: true,
					get: function () {
						return m[k]
					},
				})
		  }
		: function (o, m, k, k2) {
				if (k2 === undefined) k2 = k
				o[k2] = m[k]
		  })
var __setModuleDefault =
	(this && this.__setModuleDefault) ||
	(Object.create
		? function (o, v) {
				Object.defineProperty(o, 'default', { enumerable: true, value: v })
		  }
		: function (o, v) {
				o['default'] = v
		  })
var __importStar =
	(this && this.__importStar) ||
	function (mod) {
		if (mod && mod.__esModule) return mod
		var result = {}
		if (mod != null)
			for (var k in mod)
				if (k !== 'default' && Object.hasOwnProperty.call(mod, k))
					__createBinding(result, mod, k)
		__setModuleDefault(result, mod)
		return result
	}
var __importDefault =
	(this && this.__importDefault) ||
	function (mod) {
		return mod && mod.__esModule ? mod : { default: mod }
	}
Object.defineProperty(exports, '__esModule', { value: true })
exports.make = exports.convert = void 0
/* eslint new-cap:0 */
const fdir_1 = __importDefault(require('fdir'))
const errlop_1 = __importDefault(require('errlop'))
const rimraf_1 = __importDefault(require('rimraf'))
const mkdirp_1 = __importDefault(require('mkdirp'))
const path_1 = require('path')
const fs_1 = require('fs')
const log = __importStar(require('./log.js'))
async function rimrafp(p) {
	return new Promise(function (resolve, reject) {
		rimraf_1.default(p, function (err) {
			if (err) return reject(err)
			resolve()
		})
	})
}
const { readFile, writeFile } = fs_1.promises
async function exists(p) {
	return new Promise(function (resolve) {
		fs_1.exists(p, resolve)
	})
}
async function readJSON(path) {
	return JSON.parse(await readFile(path, 'utf-8'))
}
async function writeJSON(path, data) {
	const str = JSON.stringify(data, null, '  ')
	await writeFile(path, str, 'utf-8')
}
async function ensureFile(p, data) {
	await mkdirp_1.default(path_1.dirname(p))
	return await writeFile(p, data)
}
// https://deno.land/std/node
const builtins = [
	'assert',
	'buffer',
	'child_process',
	'cluster',
	'console',
	'crypto',
	'dgram',
	'dns',
	'events',
	'fs',
	'http',
	'http2',
	'https',
	'module',
	'net',
	'os',
	'path',
	'perf_hooks',
	'process',
	'querystring',
	'readline',
	'repl',
	'stream',
	'string_decoder',
	'sys',
	'timers',
	'tls',
	'tty',
	'url',
	'util',
	'v8',
	'vm',
	'worker_threads',
	'zlib',
]
const compat = [
	'events',
	'fs',
	'module',
	'os',
	'path',
	'process',
	'querystring',
	'timers',
	'util',
]
async function convert(opts) {
	// prepare
	const { path, cache, nm, deps, devDeps } = opts
	// check
	const cachedResult = cache[path]
	if (cachedResult) {
		return cachedResult
	}
	if ((await exists(path)) === false) {
		const error = new Error(`${path}: does not exist`)
		cache[path] = error
		log.warn(error)
		return error
	}
	// get the file content
	const sourceContent = await readFile(path, 'utf-8')
	const imports = {}
	// get the imports
	const matches = sourceContent.matchAll(/ from ['"]([^'"]+)['"]/g)
	for (const match of matches) {
		// const index = match.index
		const source = match[1]
		let target = source
		let error
		// check if local dependency, if so, ensure .ts extension
		// and ensure it is supported itself
		if (source.startsWith('.')) {
			log.log(`${path}: imports internal path [${source}]`)
			// ensure extension
			if (source.endsWith('/')) {
				target += 'index.ts'
				log.log(`${path}: mapped [${source}] to [${target}]`)
			} else {
				const ext = path_1.extname(source)
				if (ext === '') {
					target = source + '.ts'
					log.log(`${path}: mapped [${source}] to [${target}]`)
				} else if (ext === '.js') {
					target = source.replace(ext, '.ts')
					log.log(`${path}: mapped [${source}] to [${target}]`)
				} else if (ext !== '.ts') {
					error = new Error(
						`${path}: imported [${source}] which had an unsupported extension of [${ext}]`
					)
					log.warn(error)
				}
			}
			// convert it premptively, to ensure it is also suitable
			if (!error) {
				const targetPath = path_1.resolve(path_1.dirname(path), target)
				try {
					await convert({ ...opts, path: targetPath })
				} catch (err) {
					error = new errlop_1.default(
						`${path}: failed to import [${source}] which we tried to resolve to [${targetPath}]`,
						err
					)
					log.warn(error)
				}
			}
			// map
			imports[source] = [match, error || target]
			continue
		}
		// check if remote depednency, if so, ignore
		if (
			source.startsWith('http:') ||
			source.startsWith('https:') ||
			source.startsWith('/')
		) {
			log.log(`${path}: imports remote path [${source}]`)
			imports[source] = [match, source]
			continue
		}
		// check if package path, or remote dependency
		let dep = source,
			entry = ''
		if (source.includes('/')) {
			const parts = source.split('/')
			dep = parts.shift()
			if (dep[0] === '@') {
				dep += '/' + parts.shift()
			}
			entry = parts.join('/')
			log.log(
				`${path}: imports suspected package [${dep}] with entry [${entry}]`
			)
			// actually continue
		}
		// check if builtin
		if (!entry && builtins.includes(dep)) {
			// otherwise, probably node builtin, check for compat
			if (compat.includes(dep)) {
				target = `https://deno.land/std/node/${dep}.ts`
				log.log(
					`${path}: imports node builtin [${dep}] with compat at [${target}]`
				)
				imports[source] = [match, target]
				continue
			}
			// fail as non-compat node builtin
			error = new Error(
				`${path}: imports node builtin [${dep}] which does not have a compat layer`
			)
			log.warn(error)
			imports[source] = [match, error]
		} else {
			// check if package, if so, check for deno entry, if so use that, otherwise use main
			const version = deps[dep] ?? devDeps[dep]
			if (version) {
				try {
					// consider fetching from unpkg rather than node_modules
					const depPkgPath = path_1.join(nm, dep, 'package.json')
					const depPkg = await readJSON(depPkgPath).catch((err) =>
						Promise.reject(
							new errlop_1.default(
								`${path}: imports a known but uninstalled dependency [${dep}]: install it to node_modules and try again`,
								err
							)
						)
					)
					const deno = depPkg?.deno
					const main = depPkg?.main
					const resolvedEntry = entry || deno || main
					const target = `https://unpkg.com/${dep}@${version}/${resolvedEntry}`
					log.log(`${path}: imports package [${target}]`)
					if (!resolvedEntry.endsWith('.ts')) {
						error = new Error(
							`${path}: imported [${source}] which had an extension other than .ts`
						)
						log.warn(error)
					}
					imports[source] = [match, error || resolvedEntry]
					continue
				} catch (error) {
					imports[source] = [match, error]
					log.warn(error)
				}
			}
			// uninstalled dependency
			error = new Error(
				`${path}: imports an unknown dependency [${dep}]: add it to package.json and try again`
			)
			log.warn(error)
			imports[source] = [match, error]
		}
	}
	// swap out the imports with deno ones
	let error
	let offset = 0
	let resultContent = sourceContent
	for (const [source, [match, result]] of Object.entries(imports)) {
		if (result instanceof Error) {
			error = new errlop_1.default(result, error)
			continue
		}
		const target = result
		const cursor = match.index + offset
		const source = match[0]
		const replacement = source.replace(match[1], target)
		resultContent =
			resultContent.substring(0, cursor) +
			replacement +
			resultContent.substring(cursor + source.length)
		offset += replacement.length - source.length
	}
	// return
	if (error) {
		log.warn(`${path}: has incompatibilities:\n`, error.stack)
		cache[path] = error
		return error
	} else {
		log.success(`${path}: converted to deno successfully`)
		cache[path] = resultContent
		return resultContent
	}
}
exports.convert = convert
async function make({
	cwd = process.cwd(),
	failOnEntryIncompatibility = true,
	failOnTestIncompatibility = false,
	failOnOtherIncompatibility = false,
} = {}) {
	// paths
	const pkgPath = path_1.join(cwd, 'package.json')
	const pkg = await readJSON(pkgPath).catch((err) =>
		Promise.reject(
			new errlop_1.default('require package.json file to be present', err)
		)
	)
	// prepare
	const denoEditionDirectory = 'edition-deno'
	const denoEditionPath = path_1.join(cwd, denoEditionDirectory)
	const nm = path_1.join(cwd, 'node_modules')
	const deps = pkg.dependencies || {}
	const devDeps = pkg.devDependencies || {}
	// check editions
	const sourceEdition = pkg?.editions && pkg.editions[0]
	if (
		!sourceEdition ||
		!sourceEdition.tags?.includes('typescript') ||
		!sourceEdition.tags?.includes('import')
	) {
		throw new Error(
			'package.json require typescript source edition that uses ESM, refer to https://editions.bevry.me'
		)
	}
	// get the ts files of the source directory
	const sourceEditionPath = path_1.join(cwd, pkg.editions[0].directory)
	const api = new fdir_1.default()
		.filter((path) => path.endsWith('.ts'))
		.crawl(sourceEditionPath)
	const filenames = await api.withPromise()
	// delete the old files
	await rimrafp(denoEditionPath)
	// convert all the files
	let fail
	const files = {}
	for (const filename of filenames) {
		const path = path_1.join(sourceEditionPath, filename)
		const result = await convert({ path, cache: files, nm, deps, devDeps })
		if (result instanceof Error) {
			if (filename === sourceEdition.entry) {
				// entry
				if (failOnEntryIncompatibility) {
					const error = new errlop_1.default(
						`${path}: has entry incompatibilities`,
						result
					)
					log.error(error)
					fail = new errlop_1.default(error, fail)
				} else {
					const error = new errlop_1.default(
						`${path}: has entry incompatibilities, which we have configured to ignore`,
						result
					)
					log.special(error)
				}
			} else if (filename.includes('test')) {
				// test
				if (failOnTestIncompatibility) {
					const error = new errlop_1.default(
						`${path}: has test incompatibilities`,
						result
					)
					log.error(error)
					fail = new errlop_1.default(error, fail)
				} else {
					const error = new errlop_1.default(
						`${path}: has test incompatibilities, which we have configured to ignore`,
						result
					)
					log.special(error)
				}
			} else {
				// other
				// eslint-disable-next-line no-lonely-if
				if (failOnOtherIncompatibility) {
					const error = new errlop_1.default(
						`${path}: has other incompatibilities`,
						result
					)
					log.error(error)
					fail = new errlop_1.default(error, fail)
				} else {
					const error = new errlop_1.default(
						`${path}: has other incompatibilities, which we have configured to ignore`,
						result
					)
					log.special(error)
				}
			}
		} else {
			// write the new files
			const denoPath = path_1.join(denoEditionPath, filename)
			log.status(`writing [${denoPath}]...`)
			await ensureFile(denoPath, result)
			log.status(`...wrote [${denoPath}]`)
		}
	}
	// delete deno edition, will be re-added later if it is suitable
	pkg.editions = pkg.editions.filter(
		(e) => e.directory !== denoEditionDirectory
	)
	// failure
	if (fail) {
		// delete deno entry
		delete pkg.deno
		// save
		writeJSON(pkgPath, pkg)
		// return
		return Promise.reject(fail)
	} else {
		// success
		// add deno edition
		const denoEdition = {
			description: 'TypeScript source code made to be compatible with Deno',
			directory: denoEditionDirectory,
			entry: sourceEdition.entry,
			tags: ['typescript', 'import', 'deno'],
			engines: {
				deno: true,
				browsers: Boolean(pkg.browser),
			},
		}
		pkg.editions.push(denoEdition)
		// add deno entry
		pkg.deno = path_1.join(denoEdition.directory, denoEdition.entry)
		// save
		writeJSON(pkgPath, pkg)
		// return
		return true
	}
}
exports.make = make
